#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <math.h>
#include <vector>
#include <string>
#include "main.h"
#include "Command.h"
#include "Vehicle.h"
#include "Level.h"
#include "GameState.h"
#include "fontsys.h"


using namespace std;

typedef vector<string> Tokens;


static void tokenize(const string &str, Tokens &tokens,
                     const string &delimiters = " ")
{
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    string::size_type pos = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        lastPos = str.find_first_not_of(delimiters, pos);
        pos = str.find_first_of(delimiters, lastPos);
    }
}


static void initializeVehicleCommands(Vehicle &vehicle, char *filename)
{
    FILE *fp;
    INT length;
    char *wholeFile;

    fp = fopen(filename, "rb");
    if (fp == NULL)
    {
        errorMessage("File error", "Cannot open %s", filename);
        exit(EXIT_FAILURE);
    }
    fseek(fp, 0, SEEK_END);
    length = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    wholeFile = new char[length + 1];
    fread(wholeFile, length, 1, fp);
    wholeFile[length] = 0;

    fclose(fp);

    vehicle.clearCommands();

    Tokens lines;
    string wholeFileString(wholeFile);
    tokenize(wholeFileString, lines, "\r\n");

    Tokens::iterator it = lines.begin();
    Tokens::iterator itend = lines.end();
    for (; it != itend; ++it)
    {
        const string &line = *it;
        if (line.length() > 0 && line[0] == '#')
            continue;

        Tokens tokens;
        tokenize(line, tokens, " \t");
        if (tokens.size() == 0)
            continue;

        Command *command = NULL;
        if (tokens[0] == "wait")
        {
            bool ok = false;
            if (tokens.size() >= 2)
            {
                INT32 time = atoi(tokens[1].c_str());
                if (time > 0)
                {
                    command = WaitCommand::create(time);
                    ok = true;
                }
            }
            if (!ok)
                errorMessage("Error", "Invalid wait command: %s", line.c_str());
        }
        else if (tokens[0] == "forward")
            command = ForwardCommand::create();
        else if (tokens[0] == "release")
            command = ReleaseCommand::create();
        else if (tokens[0] == "reverse")
            command = ReverseCommand::create();
        else if (tokens[0] == "brake")
            command = BrakeCommand::create();
        else if (tokens[0] == "left")
            command = LeftCommand::create();
        else if (tokens[0] == "straight")
            command = StraightCommand::create();
        else if (tokens[0] == "right")
            command = RightCommand::create();

        if (command == NULL)
            errorMessage("Error", "Invalid command: %s", line.c_str());
        else
            vehicle.addCommand(command);
    }
}


void GameState::setLevelNumber(INT levelNumber)
{
    mLevelNum = levelNumber;
}


void GameState::onActivate(UINT32 tick)
{
    mLevel.load(mLevelNum);

    mVehicle.setImage(loadImage("data/vehicle.png"));

    mFont = new Font("data/font.png");

    reset(tick);
}


void GameState::reset(UINT32 tick)
{
    initializeVehicleCommands(mVehicle, mFilename);
    mVehicle.reset(0, mLevel.getStartX(), mLevel.getStartY(),
                   mLevel.getStartAngle());
    mStartTick = tick;
}


void GameState::onDeactivate()
{
    delete mFont;
    mFont = NULL;
}


INT GameState::update(UINT32 tick, KeyEventList &keyEvents)
{
    mTick = tick - mStartTick;
    mVehicle.update(mTick, mLevel);

    if (keyEvents.size() > 0)
    {
        SDLKey key = *keyEvents.begin();
        keyEvents.pop_front();

        switch (key)
        {
        case SDLK_SPACE:
            reset(tick);
            break;
        }
    }


    return 1;
}


void GameState::drawTime()
{
    char str[256];
    UINT32 time = mVehicle.getLevelDone() ?
                  mVehicle.getLevelDoneTick() :
                  mTick;
    INT x = WIDTH - 32, y = HEIGHT - 16;

    mFont->drawString(x - 43, y + 2,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      ":", 0, 0, 0);
    mFont->drawString(x - 13, y + 2,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      ".", 0, 0, 0);
    mFont->drawString(x - 45, y,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM, ":");
    mFont->drawString(x - 15, y,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM, ".");

    sprintf(str, "%02d", (time / 10) % 100);
    mFont->drawString(x + 2, y + 2,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      str, 0, 0, 0);
    mFont->drawString(x, y,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM, str);
    sprintf(str, "%02d", (time / 1000) % 100);
    mFont->drawString(x - 28, y + 2,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      str, 0, 0, 0);
    mFont->drawString(x - 30, y,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM, str);
    sprintf(str, "%02d", time / 60000);
    mFont->drawString(x - 58, y + 2,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      str, 0, 0, 0);
    mFont->drawString(x - 60, y,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM, str);
}


void GameState::drawLogo()
{
    const char title[] = "Script Kiddie Karting " VERSION_STR;
    INT x, y = 4;
    INT spacing = mFont->getCharacterSpacing();
    UINT32 t = mTick;
    FLOAT r = sin(t * 3e-3) * 0.2 + 0.8;
    FLOAT g = sin(t * 4e-4) * 0.2 + 0.8;
    FLOAT b = sin(t * 5e-4) * 0.2 + 0.8;
    mFont->setCharacterSpacing(1);
    x = 16 + mFont->getStringWidth(title) / 2;
    mFont->drawString(x, y, FONT_ALIGN_HORIZ_CENTER,
                      title, r * 0.2f, g * 0.2f, b * 0.2f, 0.25f);
    mFont->setCharacterSpacing(0);
    mFont->drawString(x, y, FONT_ALIGN_HORIZ_CENTER,
                      title, r * 0.4f, g * 0.4f, b * 0.4f, 0.5f);
    mFont->setCharacterSpacing(-1);
    mFont->drawString(x, y, FONT_ALIGN_HORIZ_CENTER,
                      title, r, g, b);
    mFont->setCharacterSpacing(spacing);
}


void GameState::drawLevelInfo()
{
    char str[256];

    sprintf(str, "Level %d", mLevelNum);
    mFont->drawString(WIDTH - 16, 10,
                      FONT_ALIGN_RIGHT, str, 1, 1, 1, 0.5f);

    mFont->drawString(WIDTH - 16, 10 + mFont->getFontHeight(),
                      FONT_ALIGN_RIGHT, mFilename, 1, 1, 1, 0.5f);

    mFont->drawString(WIDTH / 2, HEIGHT - 8,
                      FONT_ALIGN_HORIZ_CENTER | FONT_ALIGN_BOTTOM,
                      "Press space to reload and restart",
                      1, 1, 1, 0.5f);
}


void GameState::drawStatus()
{
    if (mVehicle.getLevelDone())
        mFont->drawString(32, HEIGHT - 16, FONT_ALIGN_BOTTOM, "Finished!",
                          0.5f, 1, 0.5f);
    else
    {
        char str[256];
        if (mVehicle.getCheckpointPassed() < mLevel.getCheckpoints())
            sprintf(str, "Checkpoint %d", mVehicle.getCheckpointPassed() + 1);
        else
            sprintf(str, "Goal");
        mFont->drawString(32, HEIGHT - 16 - mFont->getFontHeight(),
                          FONT_ALIGN_BOTTOM, "Destination:", 0, 0, 0, 0.3f);
        mFont->drawString(34, HEIGHT - 14, FONT_ALIGN_BOTTOM, str,
                          0, 0, 0, 0.3f);
        mFont->drawString(32, HEIGHT - 16, FONT_ALIGN_BOTTOM, str,
                          1, 0.5f, 0.5f);
    }
}


void GameState::render(SDL_Surface *screen)
{
    enter2DMode(WIDTH, HEIGHT);

    mLevel.render();

    mVehicle.render();

    drawTime();
    drawLogo();
    drawLevelInfo();
    drawStatus();

    leave2DMode();
}
