#include <math.h>
#include "SDL_image.h"
#include "main.h"
#include "Vehicle.h"
#include "Level.h"


#define HITDETECT_POINTS    16


Level::Level() : mMask(NULL)
{
}


Level::~Level()
{
    freeImage(mImage);
    delete[] mMask;
    mMask = NULL;
}


void Level::load(INT levelNumber)
{
    CHAR filename[256];

    sprintf(filename, "data/level%02dimage.jpg", levelNumber);
    mImage = loadImage(filename);

    sprintf(filename, "data/level%02dmask.png", levelNumber);
    SDL_Surface *mask = IMG_Load(filename);
    if (mask == NULL)
    {
        errorMessage("Error", "Cannot load level %d mask file %s",
                     levelNumber, filename);
        exit(EXIT_FAILURE);
    }

    if (mask->format->BitsPerPixel <= 8)
    {
        errorMessage("Error", "Level %d mask file %s bits per pixel <= 8",
                     levelNumber, filename);
        exit(EXIT_FAILURE);
    }
    if (mask->w != mImage.w || mask->h != mImage.h)
    {
        errorMessage("Error", "Level mask must be same size with level image!"
                     "\nLevel %d (%s) image: %dx%d, mask: %dx%d",
                     levelNumber, filename,
                     mImage.w, mImage.h, mask->w, mask->h);
        exit(EXIT_FAILURE);
    }
    if (SDL_LockSurface(mask) < 0)
    {
        errorMessage("Error", "Cannot lock mask surface");
        exit(EXIT_FAILURE);
    }
    delete[] mMask;
    mWidth = mask->w;
    mHeight = mask->h;
    mMask = new UINT8[mWidth * mHeight];

    mCheckpoints = 0;
    UINT32 o, lineO = 0, maskO = 0;
    INT x, y;
    for (y = 0; y < mask->h; ++y)
    {
        o = lineO;
        for (x = 0; x < mask->w; ++x)
        {
            o += mask->format->BytesPerPixel;
            UINT32 pixel = *((UINT32 *)((UINT8 *)mask->pixels + o));
            Uint8 red, green, blue;
            SDL_GetRGB(pixel, mask->format, &red, &green, &blue);
            UINT8 r = (red >> 7) & 1;
            UINT8 g = (green >> 7) & 1;
            UINT8 b = (blue >> 7) & 1;
            UINT8 col = (r << 2) | (g << 1) | b;

            switch (col)
            {
            case LMU_STARTPOINT:
                mStartX = x;
                mStartY = y;
                col = LMU_ROAD;
                break;
            case LMU_CHECKPOINT1:
                if (mCheckpoints < 1)
                    mCheckpoints = 1;
                break;
            case LMU_CHECKPOINT2:
                if (mCheckpoints < 2)
                    mCheckpoints = 2;
                break;
            case LMU_CHECKPOINT3:
                if (mCheckpoints < 3)
                    mCheckpoints = 3;
                break;
            case LMU_GOAL:
                break;
            case LMU_STARTDIRECTION:
                mStartDirX = x;
                mStartDirY = y;
                col = LMU_ROAD;
                break;
            default:
                break;
            }

            mMask[maskO++] = col;   // maps directly to LEVEL_MASK_UNIT
        }
        lineO += mask->pitch;
    }

    SDL_UnlockSurface(mask);

    mStartAngle = atan2(mStartY - mStartDirY, mStartX - mStartDirX);
}


void Level::update(UINT32 tick)
{
}

    
void Level::render()
{
    drawImage(mImage);
}


void Level::hitDetect(Vehicle *vehicle)
{
    FLOAT radius = vehicle->getVehicleHitDetectRadius();
    FLOAT vx, vy, angle = 0, angleStep = 2 * (FLOAT)PI / HITDETECT_POINTS;
    INT a;

    vehicle->getPosition(vx, vy);

    for (a = 0; a < HITDETECT_POINTS; ++a, angle += angleStep)
    {
        INT x = (INT)clamp(vx + (FLOAT)cos(angle) * radius, 0, WIDTH - 1);
        INT y = (INT)clamp(vy + (FLOAT)sin(angle) * radius, 0, HEIGHT - 1);

        switch (mMask[y * mWidth + x])
        {
        case LMU_CHECKPOINT1:
            vehicle->hitCheckpoint(1);
            break;
        case LMU_CHECKPOINT2:
            vehicle->hitCheckpoint(2);
            break;
        case LMU_CHECKPOINT3:
            vehicle->hitCheckpoint(3);
            break;
        case LMU_GOAL:
            vehicle->hitGoal(mCheckpoints);
            break;
        case LMU_WALL:
            vehicle->hitWall();
            break;
        }
    }
}
