#include <math.h>
#include <assert.h>
#include <list>
#include "main.h"
#include "Vehicle.h"
#include "Command.h"
#include "Level.h"


using namespace std;


static FLOAT boxstep(FLOAT x, FLOAT a, FLOAT b)
{
    return clamp((x - a) / (b - a), 0, 1);
}


Vehicle::Vehicle() :
    mVehicleMaxSpeed(5.0f),
    mVehicleAcceleration(0.02f),
    mVehicleBrakeRate(0.97f),
    mVehicleTurningRate(1.0f * (FLOAT)PI/180),
    mVehicleHitDetectRadius(24)
{
}


Vehicle::~Vehicle()
{
    clearCommands();
}


void Vehicle::wait(INT time)
{
    mNextCommandTick += time;
}


void Vehicle::forward()
{
    mAcceleration = 1;
    mBraking = 0;
}


void Vehicle::release()
{
    mAcceleration = 0;
    mBraking = 0;
}


void Vehicle::reverse()
{
    mAcceleration = -1;
    mBraking = 0;
}


void Vehicle::brake()
{
    mAcceleration = 0;
    mBraking = 1;
}


void Vehicle::left()
{
    mTurning = -1;
}


void Vehicle::straight()
{
    mTurning = 0;
}


void Vehicle::right()
{
    mTurning = 1;
}


void Vehicle::reset(UINT32 tick, FLOAT x, FLOAT y, FLOAT angle)
{
    mNextCommandTick = tick;
    mX = x;
    mY = y;
    mAngle = angle;
    mCurrentCommand = mCommandList.begin();
    mSpeed = 0;
    mAngle = angle;
    mTurning = mAcceleration = mBraking = 0;
    mCheckpointPassed = 0;
    mLevelDone = false;
    mLevelDoneTick = 0;
}


void Vehicle::update(UINT32 tick, Level &level)
{
    mCurrentTick = tick;

    mOldX = mX;
    mOldY = mY;
    mOldAngle = mAngle;

    while (mNextCommandTick <= tick &&
           mCurrentCommand != mCommandList.end())
    {
        (*mCurrentCommand)->apply(this);
        ++mCurrentCommand;
    }

    mSpeed += mAcceleration * mVehicleAcceleration;
    if (mSpeed > mVehicleMaxSpeed)
        mSpeed = mVehicleMaxSpeed;
    if (mSpeed < -mVehicleMaxSpeed)
        mSpeed = -mVehicleMaxSpeed;
    if (mBraking)
        mSpeed *= mVehicleBrakeRate;
    mAngle += (mSpeed < 0 ? -1 : 1) * mVehicleTurningRate * mTurning *
              boxstep(fabs(mSpeed / mVehicleMaxSpeed), 0.01f, 0.1f);
    mX += (FLOAT)cos(mAngle) * mSpeed;
    mY += (FLOAT)sin(mAngle) * mSpeed;

    level.hitDetect(this);
}


void Vehicle::hitCheckpoint(INT num)
{
    if (mCheckpointPassed == num - 1)
        mCheckpointPassed = num;
}


void Vehicle::hitGoal(INT checkpoints)
{
    if (mCheckpointPassed == checkpoints && !mLevelDone)
    {
        mBraking = 1;
        mAcceleration = 0;
        mCurrentCommand = mCommandList.end();
        mLevelDone = true;
        mLevelDoneTick = mCurrentTick;
    }
}


void Vehicle::hitWall()
{
    mX = mOldX;
    mY = mOldY;
    mAngle = mOldAngle;
    mSpeed = 0;
    mBraking = 1;
}


void Vehicle::render()
{
    drawImageRotatedCentered(mImage, mX, mY, mAngle + PI / 2);
}


void Vehicle::clearCommands()
{
    CommandList::iterator it = mCommandList.begin();
    CommandList::iterator itend = mCommandList.end();
    for (; it != itend; ++it)
        delete *it;
    mCommandList.clear();
}


void Vehicle::addCommand(Command *command)
{
    assert(command != NULL);
    mCommandList.push_back(command); 
}


void Vehicle::setImage(const IMAGE &image)
{
    mImage = image;
}
