#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "main.h"
#include "Command.h"
#include "Vehicle.h"
#include "Level.h"
#include "GameState.h"


static SDL_Surface *screen;
static INT appAlive = 1;
static UINT32 appTime, appStartTime;
static UINT32 nextUpdateTime;

static KeyEventList keyEvents;
static UINT8 *keyState;

static State *currentState;
static GameState *gameState;


UINT32 getAppTime()
{
    return appTime;
}


static int eventFilter(const SDL_Event *event)
{
    SDLKey keysym;

    switch (event->type)
    {
    case SDL_KEYDOWN:
        keyState = SDL_GetKeyState(NULL);

        if (keyState[SDLK_ESCAPE] == SDL_PRESSED) {
            appAlive = 0;
            return 0;
        }
        keysym = event->key.keysym.sym;
        switch (keysym)
        {
        case SDLK_KP4:
            keyEvents.push_back(SDLK_LEFT);
            break;
        case SDLK_KP6:
            keyEvents.push_back(SDLK_RIGHT);
            break;
        case SDLK_KP8:
            keyEvents.push_back(SDLK_UP);
            break;
        case SDLK_KP2:
            keyEvents.push_back(SDLK_DOWN);
            break;
        default:
            keyEvents.push_back(keysym);
            break;
        }
        break;

    case SDL_KEYUP:
        keyState = SDL_GetKeyState(NULL);
        break;
    }
    return 1;
}


static SDL_Surface * setVideoMode(int width, int height, int bpp)
{
    Uint32 videoFlags = 0;
    
    int rSize, gSize, bSize;
    
    switch (bpp)
    {
    case 8:
        rSize = 2;
        gSize = 3;
        bSize = 3;
    case 15:
    case 16:
        rSize = 5;
        gSize = 5;
        bSize = 5;
        break;
    default:
        rSize = 8;
        gSize = 8;
        bSize = 8;
        break;
    }
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, rSize);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, gSize);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, bSize);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    videoFlags |= SDL_HWSURFACE;
    videoFlags |= SDL_DOUBLEBUF;
    videoFlags |= SDL_ANYFORMAT;
    videoFlags |= SDL_OPENGLBLIT;

    return SDL_SetVideoMode(width, height, 32, videoFlags);
}


static void resetTime()
{
    nextUpdateTime = 0;
    appStartTime = SDL_GetTicks();
    appTime = 0;
}


static char update(State *state, UINT32 currentTime)
{
    if (currentTime < nextUpdateTime)
        return 0;

    while (nextUpdateTime < currentTime)
    {
        nextUpdateTime += 1000 / state->getUpdateFPS();
    
        state->update(nextUpdateTime, keyEvents);
    }

    return 1;
}


static void initStates(int argc, char *argv[])
{
    gameState = new GameState(argv[2]);
    gameState->setLevelNumber(atoi(argv[1]));
    currentState = NULL;
}


void setMainState(MAINSTATE state)
{
    if (currentState != NULL)
        currentState->onDeactivate();

    switch (state)
    {
    case MS_GAME:
        currentState = gameState;
        break;
    }

    assert(currentState);
    currentState->onActivate(appTime);
}


static int fileExists(char *filename)
{
    FILE *fp;
    fp = fopen(filename, "r");
    if (fp != NULL)
    {
        fclose(fp);
        return 1;
    }
    return 0;
}


int main(int argc, char *argv[])
{
    if (argc < 3 || !fileExists(argv[2]))
    {
        errorMessage("Startup Error",
                     "Give the level number as the first parameter"
                     "and the input script as the second parameter!\n\n"
                        "Example:\n\t%s 1 script01.txt",
                     argv[0]);
        exit(EXIT_FAILURE);
    }

    if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE) < 0)
    {
        fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
    atexit(SDL_Quit);

    {
        char caption[512];
        sprintf(caption, "%s - Script Kiddie Karting", argv[2]);
        SDL_WM_SetCaption(caption, "");
    }

    if ((screen = setVideoMode(WIDTH, HEIGHT, 32)) == NULL &&
        (screen = setVideoMode(WIDTH, HEIGHT, 24)) == NULL
        (screen = setVideoMode(WIDTH, HEIGHT, 16)) == NULL
        (screen = setVideoMode(WIDTH, HEIGHT, 8)) == NULL)
    {
        fprintf(stderr, "Unable to set %dx%d mode: %s\n",
                WIDTH, HEIGHT, SDL_GetError());
        exit(1);
    }

    SDL_SetEventFilter(eventFilter);

    resetTime();

    srand(time(NULL));


    initStates(argc, argv);

    setMainState(MS_GAME);


    while (appAlive && !SDL_QuitRequested())
    {
        char updated = 0;
        unsigned int time = SDL_GetTicks() - appStartTime;
        appTime = BLURMIX(appTime, time, 4);

        updated |= update(currentState, appTime);

        if (updated)
        {
            currentState->render(screen);
            //SDL_Flip(screen);
            glFlush();
            SDL_GL_SwapBuffers();
        }
        else
        {
            SDL_Delay(1);
        }
    }

    return EXIT_SUCCESS;
}
